package com.ibm.rcp.security.icard;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.URL;
import java.net.UnknownHostException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.higgins.rp.icard.token.identity.DigitalIdentity;
import org.eclipse.higgins.sts.api.IElement;
import org.eclipse.higgins.sts.common.Constants;
import org.eclipse.higgins.sts.spi.IXMLSecurityExtension;
import org.eclipse.higgins.sts.xmlsecurity.apache.XMLSecurityApacheExtension;
import org.osgi.framework.Bundle;

import com.ibm.rcp.security.auth.AuthProperties;

public class ICardLoginModule implements LoginModule {
	private Subject subject;
	private KeyStore keystore;
	
	private String keyAlias = "leaf";
	private String keyPass = "changeit";
	
	public boolean abort() throws LoginException {
		return true;
	}

	public boolean commit() throws LoginException {
		return true;
	}

	public void initialize(Subject subject, CallbackHandler callbackHandler,
			Map sharedState, Map options) {
		this.subject = subject;
		this.keystore = getKeyStore();
	}
	
	private KeyStore getKeyStore() {		
		KeyStore ks;
		try {
			ks = KeyStore.getInstance("JKS");
		} catch (KeyStoreException e1) {
			e1.printStackTrace();
			return null;
		}
		Bundle bundle = Platform.getBundle("com.ibm.rcp.security.icard");
		URL rootURL = FileLocator.find(bundle, new Path("/"), null);
		URL fileURL = null;
		try {
			fileURL = FileLocator.resolve(rootURL);
		} catch (IOException e1) {
			e1.printStackTrace();
			return null;
		}
		String absolutePath = fileURL.getPath();
		FileInputStream fis;
		try {
			fis = new FileInputStream(new File(absolutePath + "/localhost.jks"));
			ks.load(fis, keyPass.toCharArray());		
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
		
		return ks;
	}

	public boolean login() throws LoginException {
		Set publicSet = subject.getPublicCredentials(AuthProperties.class);
		AuthProperties authProperties = null;
		if ((null != publicSet) && 1 == publicSet.size()) {
			authProperties = (AuthProperties) publicSet.toArray()[0];
			String message = createHBXRequest(authProperties);
			String token = sendHBXMessage(message);
			DigitalIdentity di = decryptToken(token);
			if(di != null){
				subject.getPrivateCredentials().add(di);
				return true;
			}
		}
		return false;
	}
	
	private DigitalIdentity decryptToken(String strToken) throws LoginException{
		try	{
			org.eclipse.higgins.sts.api.IElement elemToken = new org.eclipse.higgins.sts.common.Element();
			elemToken.set(strToken);
			IXMLSecurityExtension secext = new XMLSecurityApacheExtension(); 
			IElement ie = null;
			secext.configure(new HashMap(), null, new HashMap());

			PrivateKey key = (PrivateKey)keystore.getKey(keyAlias, keyPass.toCharArray());
			System.out.println("Decrypt token using key " + key + " key algorithm " + key.getAlgorithm());
			ie = secext.DecryptElement(elemToken, (PrivateKey)(keystore.getKey("leaf", "changeit".toCharArray())));
			System.out.println("Decrypted token looks like\n" + ie.getAs(java.lang.String.class));
			// verify that it's signed correctly
			secext.VerifyEnveloped(ie, new Constants());
			DigitalIdentity di = DigitalIdentity.fromXml(ie);
			System.out.println(di);
			return di;
		} catch (Exception e)	{
			LoginException le =  new LoginException("can not decrypt the token");
			le.initCause(e);
			throw le;
		}
	}
	
	private String sendHBXMessage(String message) throws LoginException{
		try {
			Socket socket = new Socket("localhost", 8088);
			PrintWriter out = new PrintWriter(socket.getOutputStream(), true);
			BufferedReader in = new BufferedReader(
					new InputStreamReader(socket.getInputStream()));
			out.print(message);
			out.flush();
			out.print("\0");
			out.flush();
						
			StringBuffer sb = new StringBuffer();
			while (true) {
				int ch = in.read();
				if (ch == -1  || ch == '\0' || ch== 0){
					break; 						
				} else {
					sb.append((char) (ch & 0xFFFF));
				}
			}
			return sb.toString();
		} catch (UnknownHostException e) {
			LoginException le =  new LoginException("Unknown host name");
			le.initCause(e);
			throw le;
		} catch (IOException e) {
			LoginException le =  new LoginException("");
			le.initCause(e);
			throw le;
		}
	}
	
	private String createHBXRequest(AuthProperties authProperties) 
	throws LoginException{		
		String url = authProperties.getProperty(AuthProperties.SERVER);
		String tokenType = authProperties.getProperty("tokenType");
		String requiredClaims = authProperties.getProperty("requiredClaims");
		String optionalClaims = authProperties.getProperty("optionalClaims");
		String issuer = authProperties.getProperty("issuer");
		String issuerPolicy = authProperties.getProperty("issuerPolicy");
		String privacyUrl = authProperties.getProperty("privacyUrl");
		String privacyVersion = authProperties.getProperty("privacyVersion");	
		
		String cert = null;
		
		try {
			X509Certificate x509cert = (X509Certificate) this.keystore.getCertificate("leaf");
			byte[] certBytes = x509cert.getEncoded();
			cert = org.apache.xml.security.utils.Base64.encode(certBytes);
		} catch (Exception e) {
			LoginException le =  new LoginException("Can't get XPD's certificate.");
			le.initCause(e);
			throw le;
		} 
		
		StringBuffer sb = new StringBuffer();
		sb.append("<hbx_request>");
		sb.append("<document_URL>").append(url).append("</document_URL>");
		sb.append("<parameters>");
		if(tokenType!=null && tokenType.length()>0)
			sb.append("<parameter name=\"TokenType\" value=\"" + tokenType + "\"/>");
		if(requiredClaims!=null && requiredClaims.length()>0)
			sb.append("<parameter name=\"requiredClaims\" value=\"" + requiredClaims +  "\"/>");
		if(optionalClaims!=null && optionalClaims.length()>0)
			sb.append("<parameter name=\"optionalClaims\" value=\"" + optionalClaims +  "\"/>");
		if(issuer!=null && issuer.length()>0)
			sb.append("<parameter name=\"issuer\" value=\"" + issuer +  "\"/>");
		if(issuerPolicy!=null && issuerPolicy.length()>0)
			sb.append("<parameter name=\"issuerPolicy\" value=\"" + issuerPolicy +  "\"/>");
		if(privacyUrl!=null && privacyUrl.length()>0)
			sb.append("<parameter name=\"privacyUrl\" value=\"" + privacyUrl +  "\"/>");
		if(privacyVersion!=null && privacyVersion.length()>0)
			sb.append("<parameter name=\"privacyVersion\" value=\"" + privacyVersion +  "\"/>");
		sb.append("</parameters>");
		sb.append("<certificate_chain>");
		sb.append("<certificate>");
		sb.append(cert);
		sb.append("</certificate>");
		sb.append("</certificate_chain>");
		sb.append("</hbx_request>");
		
		return sb.toString();
	}
	
	public boolean logout() throws LoginException {
		return true;
	}
}
